# -*- coding: utf-8 -*-
# @Time    : 7/26/21 7:18 PM
# @Author  : Jingnan
# @Email   : jiajingnan2222@gmail.com
import torch
from torch import nn
import kd_med.resnet3d as resnet

def generate_model(opt):
    """

    :param opt: require following values: model, model_depth, input_W, input_H, input_D,
    resnet_shortcut, no_cuda, n_seg_classes, phase, pretrain_path
    :return:
    """
    assert opt.model in [
        'resnet'
    ]

    if opt.model == 'resnet':
        assert opt.model_depth in [10, 18, 34, 50, 101, 152, 200]

        if opt.model_depth == 10:
            model = resnet.resnet10(
                shortcut_type=opt.resnet_shortcut,
                no_cuda=opt.no_cuda,
                num_seg_classes=opt.n_seg_classes)
        elif opt.model_depth == 18:
            model = resnet.resnet18(
                shortcut_type=opt.resnet_shortcut,
                no_cuda=opt.no_cuda,
                num_seg_classes=opt.n_seg_classes)
        elif opt.model_depth == 34:
            model = resnet.resnet34(
                shortcut_type=opt.resnet_shortcut,
                no_cuda=opt.no_cuda,
                num_seg_classes=opt.n_seg_classes)
        elif opt.model_depth == 50:
            model = resnet.resnet50(
                shortcut_type=opt.resnet_shortcut,
                no_cuda=opt.no_cuda,
                num_seg_classes=opt.n_seg_classes)
        elif opt.model_depth == 101:
            model = resnet.resnet101(
                shortcut_type=opt.resnet_shortcut,
                no_cuda=opt.no_cuda,
                num_seg_classes=opt.n_seg_classes)
        elif opt.model_depth == 152:
            model = resnet.resnet152(
                shortcut_type=opt.resnet_shortcut,
                no_cuda=opt.no_cuda,
                num_seg_classes=opt.n_seg_classes)
        elif opt.model_depth == 200:
            model = resnet.resnet200(
                shortcut_type=opt.resnet_shortcut,
                no_cuda=opt.no_cuda,
                num_seg_classes=opt.n_seg_classes)

    # if not opt.no_cuda:
    #     if len(opt.gpu_id) > 1:
    #         model = model.cuda()
    #         model = nn.DataParallel(model, device_ids=opt.gpu_id)
    #         net_dict = model.state_dict()
    #     else:
    #         import os
    #         os.environ["CUDA_VISIBLE_DEVICES"] = str(opt.gpu_id[0])
    #         model = model.cuda()
    #         model = nn.DataParallel(model, device_ids=None)
    #         net_dict = model.state_dict()
    # else:
    #     net_dict = model.state_dict()

    # # load pretrain
    # if opt.pretrain_path:
    #     print('loading pretrained model {}'.format(opt.pretrain_path))
    #     pretrain = torch.load(opt.pretrain_path)
    #     pretrain_dict = {k: v for k, v in pretrain['state_dict'].items() if k in net_dict.keys()}
    #
    #     net_dict.update(pretrain_dict)
    #     model.load_state_dict(net_dict)


    return model



# -*- coding: utf-8 -*-
# @Time    : 7/26/21 6:59 PM
# @Author  : Jingnan
# @Email   : jiajingnan2222@gmail.com
import math

import torch
import torch.nn as nn
from kd_med.pre_trained_enc import PreTrainedEnc


class EncPlusConv(nn.Module):
    """
    the out_chn of enc_s must be equal to the in_chn of conv.
    """
    def __init__(self, enc_s, conv):
        super().__init__()
        self.enc_s = enc_s
        self.conv = conv

    def forward(self, x):
        out = self.enc_s(x)
        out = self.conv(out)
        return out


class GetEncSConv:
    """
    Using class method to make sure the network only be created once and reusable many times.
    """
    enc_t = None
    enc_s = None
    enc_plus_conv = None
    chn_t = None
    chn_s = None
    dims = 3


    @classmethod
    def set_conv_config(cls, dim1_t, dim1_s):
        # o = (n - f + 2 * p) / s + 1
        # f = n - ((o - 1) * s - 2 * p)
        # p = ((o - 1) * s - n + f)/2
        if dim1_t == dim1_s:  # no need of conv at all
            cls.enc_plus_conv = cls.enc_s
            return cls.enc_plus_conv
        elif dim1_t > dim1_s:
            raise Exception(f'teacher model depth is less than student model, dim1_t: {dim1_t}, dim1_s: {dim1_s}')
        else:  # teacher model is deeper
            s = math.ceil(dim1_s / dim1_t) # down sample using stride at first, pad more if over down-sampling
            conv_sz = s + 1  # conv size should be bigger than stride
            exprected_in_size = s * (dim1_t - 1) + conv_sz
            p = math.ceil((exprected_in_size - dim1_s) / 2)
            if cls.dims == 3:
                conv = nn.Conv3d(cls.chn_s, cls.chn_t, kernel_size=(conv_sz, conv_sz, conv_sz),
                                 stride=(s, s, s), padding=p)
            else:
                conv = nn.Conv2d(cls.chn_s, cls.chn_t, kernel_size=(conv_sz, conv_sz),
                                 stride=(s, s), padding=p)
            cls.enc_plus_conv = EncPlusConv(cls.enc_s, conv)

    @classmethod
    def get(cls, enc_t: nn.Module, enc_s: nn.Module, dims: int = 3):
        """
        Chn_in = 1 always here.
        :param enc_t:
        :param enc_s:
        :param dims:
        :return:
        """
        # print(f"cls.enc_t: {cls.enc_t}, cls.enc_s: {cls.enc_s}")
        if cls.enc_t is not enc_t:  # when coming encoder is not the same as the stored one

            cls.dims = dims
            cls.enc_t = enc_t
            cls.enc_s = enc_s
            if dims == 3:
                input_tmp = torch.ones((2, 1, 128, 128, 128))
                out_t = cls.enc_t(input_tmp)
                out_s = cls.enc_s(input_tmp)
                batch_size, cls.chn_t, dim1_t, dim2_t, dim3_t = out_t.shape
                batch_size, cls.chn_s, dim1_s, dim2_s, dim3_s = out_s.shape
            else:
                input_tmp = torch.ones((2, 1, 128, 128))
                out_t = cls.enc_t(input_tmp)
                out_s = cls.enc_s(input_tmp)
                batch_size, cls.chn_t, dim1_t, dim2_t = out_t.shape
                batch_size, cls.chn_s, dim1_s, dim2_s = out_s.shape

            cls.set_conv_config(dim1_t, dim1_s)

        return cls.enc_plus_conv


def kd_loss(batch_x: torch.Tensor,
            enc_s: nn.Module,
            net_t_name: str ='resnet3d_34'):
    """
    The enc_s will share the same memory with enc_s_conv, and the enc_s_conv will be optimized by the loss of kd.
    todo: I have to put enc_s_conv to outsize otherwise the last conv will not be updated at all !!!
    :param batch_x:
    :param enc_s:
    :param net_t_name:
    :return:
    """

    if len(batch_x.shape) == 5:
        dims: int = 3
    elif len(batch_x.shape) == 4:
        dims = 2
    else:
        raise Exception(f'shape of batch_x: {batch_x.shape} is not correct')

    enc_t = PreTrainedEnc.get(net_t_name)
    # enc_s share the same memory with the enc_s in enc_s_conv, only create it once and reuse it
    enc_s_conv = GetEncSConv().get(enc_t, enc_s, dims)
    with torch.no_grad():
        out_t = enc_t(batch_x)
    out_s = enc_s_conv(batch_x)
    loss = nn.MSELoss()
    kd_loss = loss(out_t, out_s)
    return kd_loss




# -*- coding: utf-8 -*-
# @Time    : 7/28/21 3:23 PM
# @Author  : Jingnan
# @Email   : jiajingnan2222@gmail.com
import unittest
import tempfile
import os
import torch
import torch.nn as nn

from parameterized import parameterized
from kd_med.pre_trained_enc import PreTrainedEnc, UNet3DEnc
from kd_med import resnet3d
from kd_med.unet3d import UNet3D

import numpy as np


TEST_CASE_1 = ["resnet3d_10", resnet3d.resnet10(shortcut_type='B', num_seg_classes=99)]
TEST_CASE_2 = ["resnet3d_18", resnet3d.resnet18(shortcut_type='A', num_seg_classes=99)]
TEST_CASE_3 = ["resnet3d_34", resnet3d.resnet34(shortcut_type='A', num_seg_classes=99)]
TEST_CASE_4 = ["resnet3d_50", resnet3d.resnet50(shortcut_type='B', num_seg_classes=99)]
TEST_CASE_5 = ["resnet3d_101", resnet3d.resnet101(shortcut_type='B', num_seg_classes=99)]
TEST_CASE_6 = ["resnet3d_152", resnet3d.resnet152(shortcut_type='B', num_seg_classes=99)]
TEST_CASE_7 = ["resnet3d_200", resnet3d.resnet200(shortcut_type='B', num_seg_classes=99)]
TEST_CASE_8 = ["unet3d", UNet3DEnc()]


class TestPreTrainedEnc(unittest.TestCase):
    @parameterized.expand([TEST_CASE_1, TEST_CASE_2, TEST_CASE_3, TEST_CASE_4,
                           TEST_CASE_5, TEST_CASE_6, TEST_CASE_7, TEST_CASE_8])
    def test_PreTrainedEnc(self, net_name, expected_net):
        net = PreTrainedEnc().get(net_name)
        layer_names = [name for name, param in net.named_parameters()]
        expected_names = [name for name, param in expected_net.named_parameters()]
        self.assertEqual(layer_names, expected_names)

    @parameterized.expand([TEST_CASE_1, TEST_CASE_2, TEST_CASE_3, TEST_CASE_4,
                           TEST_CASE_5, TEST_CASE_6, TEST_CASE_7, TEST_CASE_8])
    def test_PreTrainedEnc_same_weights(self, net_name, expected_net):
        net1 = PreTrainedEnc().get(net_name)
        net2 = PreTrainedEnc().get(net_name)
        net3 = PreTrainedEnc().get(net_name)
        self.assertIs(net1, net2)
        self.assertIs(net1, net3)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# @Time    : 7/26/21 8:10 PM
# @Author  : Jingnan
# @Email   : jiajingnan2222@gmail.com
import torch
import torch.nn as nn
import torch.nn.functional as F


class ContBatchNorm3d(nn.modules.batchnorm._BatchNorm):
    def _check_input_dim(self, input):

        if input.dim() != 5:
            raise ValueError('expected 5D input (got {}D input)'.format(input.dim()))
        #super(ContBatchNorm3d, self)._check_input_dim(input)

    def forward(self, input):
        self._check_input_dim(input)
        return F.batch_norm(
            input, self.running_mean, self.running_var, self.weight, self.bias,
            True, self.momentum, self.eps)


class LUConv(nn.Module):
    def __init__(self, in_chan, out_chan, act):
        super(LUConv, self).__init__()
        self.conv1 = nn.Conv3d(in_chan, out_chan, kernel_size=3, padding=1)
        self.bn1 = ContBatchNorm3d(out_chan)

        if act == 'relu':
            self.activation = nn.ReLU(out_chan)
        elif act == 'prelu':
            self.activation = nn.PReLU(out_chan)
        elif act == 'elu':
            self.activation = nn.ELU(inplace=True)
        else:
            raise

    def forward(self, x):
        out = self.activation(self.bn1(self.conv1(x)))
        return out


def _make_nConv(in_channel, depth, act, double_chnnel=False):
    if double_chnnel:
        layer1 = LUConv(in_channel, 32 * (2 ** (depth+1)),act)
        layer2 = LUConv(32 * (2 ** (depth+1)), 32 * (2 ** (depth+1)),act)
    else:
        layer1 = LUConv(in_channel, 32*(2**depth),act)
        layer2 = LUConv(32*(2**depth), 32*(2**depth)*2,act)

    return nn.Sequential(layer1,layer2)


# class InputTransition(nn.Module):
#     def __init__(self, outChans, elu):
#         super(InputTransition, self).__init__()
#         self.conv1 = nn.Conv3d(1, 16, kernel_size=5, padding=2)
#         self.bn1 = ContBatchNorm3d(16)
#         self.relu1 = ELUCons(elu, 16)
#
#     def forward(self, x):
#         # do we want a PRELU here as well?
#         out = self.bn1(self.conv1(x))
#         # split input in to 16 channels
#         x16 = torch.cat((x, x, x, x, x, x, x, x,
#                          x, x, x, x, x, x, x, x), 1)
#         out = self.relu1(torch.add(out, x16))
#         return out

class DownTransition(nn.Module):
    def __init__(self, in_channel,depth, act):
        super(DownTransition, self).__init__()
        self.ops = _make_nConv(in_channel, depth,act)
        self.maxpool = nn.MaxPool3d(2)
        self.current_depth = depth

    def forward(self, x):
        if self.current_depth == 3:
            out = self.ops(x)
            out_before_pool = out
        else:
            out_before_pool = self.ops(x)
            out = self.maxpool(out_before_pool)
        return out, out_before_pool


class UpTransition(nn.Module):
    def __init__(self, inChans, outChans, depth,act):
        super(UpTransition, self).__init__()
        self.depth = depth
        self.up_conv = nn.ConvTranspose3d(inChans, outChans, kernel_size=2, stride=2)
        self.ops = _make_nConv(inChans+ outChans//2,depth, act, double_chnnel=True)

    def forward(self, x, skip_x):
        out_up_conv = self.up_conv(x)
        concat = torch.cat((out_up_conv,skip_x),1)
        out = self.ops(concat)
        return out


class OutputTransition(nn.Module):
    def __init__(self, inChans, n_labels):

        super(OutputTransition, self).__init__()
        self.final_conv = nn.Conv3d(inChans, n_labels, kernel_size=1)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        out = self.sigmoid(self.final_conv(x))
        return out


class UNet3D(nn.Module):
    # the number of convolutions in each layer corresponds
    # to what is in the actual prototxt, not the intent
    def __init__(self, n_class=1, act='relu'):
        super(UNet3D, self).__init__()


        self.up_tr256 = UpTransition(512, 512,2,act)
        self.up_tr128 = UpTransition(256,256, 1,act)
        self.up_tr64 = UpTransition(128,128,0,act)
        self.out_tr = OutputTransition(64, n_class)

        self.down_tr64 = DownTransition(1,0,act)
        self.down_tr128 = DownTransition(64,1,act)
        self.down_tr256 = DownTransition(128,2,act)
        self.down_tr512 = DownTransition(256,3,act)


    def forward(self, x):
        self.out64, self.skip_out64 = self.down_tr64(x)
        self.out128,self.skip_out128 = self.down_tr128(self.out64)
        self.out256,self.skip_out256 = self.down_tr256(self.out128)
        self.out512,self.skip_out512 = self.down_tr512(self.out256)

        self.out_up_256 = self.up_tr256(self.out512,self.skip_out256)
        self.out_up_128 = self.up_tr128(self.out_up_256, self.skip_out128)
        self.out_up_64 = self.up_tr64(self.out_up_128, self.skip_out64)
        self.out = self.out_tr(self.out_up_64)

        return self.out


# -*- coding: utf-8 -*-
# @Time    : 6/27/21 1:34 PM
# @Author  : Jingnan
# @Email   : jiajingnan2222@gmail.com

import unittest
import tempfile
import os
import torch
import torch.nn as nn

from parameterized import parameterized
from kd_med.kd_loss import EncPlusConv

import numpy as np

TEST_CASE_3d = [nn.Conv3d(1, 48, kernel_size=(3, 3, 3), stride=(2, 2, 2), padding=(3, 3, 3)),
               nn.Conv3d(48, 32, kernel_size=(3, 3, 3), stride=(2, 2, 2), padding=(3, 3, 3)),
               torch.ones((2, 1, 64, 64, 32))]

TEST_CASE_2d = [nn.Conv2d(3, 64, kernel_size=(3, 3), stride=(2, 2), padding=(3, 3)),
               nn.Conv2d(64, 32, kernel_size=(3, 3), stride=(2, 2), padding=(3, 3)),
               torch.ones((2, 3, 128, 96))]


class TestEncPlusConv(unittest.TestCase):
    @parameterized.expand([TEST_CASE_3d, TEST_CASE_2d])
    def test_EncPlusConv(self, enc, conv, input_data):

        enc_plus_conv = EncPlusConv(enc, conv)
        out_enc = enc(input_data)
        out_conv = conv(out_enc)
        out_enc_plus_conv = enc_plus_conv(input_data)
        out_1 = out_conv.clone().detach().cpu().numpy()
        out_2 = out_enc_plus_conv.clone().detach().cpu().numpy()

        self.assertIsNone(np.testing.assert_allclose(out_1, out_2, rtol=1e-5, atol=0))


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# @Time    : 7/28/21 8:34 PM
# @Author  : Jingnan
# @Email   : jiajingnan2222@gmail.com
import unittest
import tempfile
import os
import torch
import torch.nn as nn

from parameterized import parameterized
from kd_med.kd_loss import kd_loss


class Cnn2_3dEnc(nn.Module):
    def __init__(self, fc1_nodes=1024, fc2_nodes=1024, num_classes: int = 5, base: int = 8, level_node = 0):
        super().__init__()
        self.level_node = level_node
        self.features = nn.Sequential(
            nn.Conv3d(1, base, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.MaxPool3d(kernel_size=3, stride=2),
            nn.Conv3d(base, base * 2, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.MaxPool3d(kernel_size=3, stride=2),
            # nn.Conv3d(base * 2, base * 4, kernel_size=3, padding=1),
            # nn.ReLU(inplace=True),
            # nn.MaxPool3d(kernel_size=3, stride=2),
        )
        self.avgpool = nn.AdaptiveAvgPool3d((6, 6, 6))
        self.classifier = nn.Sequential(
            nn.Dropout(),
            nn.Linear(base * 2 * 6 * 6 * 6, fc1_nodes),
            nn.ReLU(inplace=True),
            nn.Dropout(),
            nn.Linear(fc1_nodes, fc2_nodes),
            nn.ReLU(inplace=True),
            nn.Dropout(),
            nn.Linear(fc2_nodes, num_classes),
        )

    def forward(self, input):  # input would be a tuple of size (1,) if only one element is input
        if self.level_node == 0:
            x = input
        else:
            x, level = input[0], input[1]

        x = self.features(x)
        # x = self.avgpool(x)
        # x = torch.flatten(x, 1)
        # x = self.classifier(x)
        return x


class Vgg11_3dEnc(nn.Module):
    def __init__(self, fc1_nodes=1024, fc2_nodes=1024, num_classes: int = 5, base: int = 8, level_node = 0):
        super().__init__()
        self.num_classes = num_classes
        self.base = base
        self.level_node = level_node

        self.features = nn.Sequential(
            nn.Conv3d(1, base, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.BatchNorm3d(base),
            nn.MaxPool3d(kernel_size=3, stride=2),

            nn.Conv3d(base, base * 2, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.BatchNorm3d(base * 2),
            nn.MaxPool3d(kernel_size=3, stride=2),

            nn.Conv3d(base * 2, base * 4, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.BatchNorm3d(base * 4),
            nn.Conv3d(base * 4, base * 4, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.BatchNorm3d(base * 4),
            nn.MaxPool3d(kernel_size=3, stride=2),

            nn.Conv3d(base * 4, base * 8, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.BatchNorm3d(base * 8),
            nn.Conv3d(base * 8, base * 8, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.BatchNorm3d(base * 8),
            nn.MaxPool3d(kernel_size=3, stride=2),

            nn.Conv3d(base * 8, base * 16, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.BatchNorm3d(base * 16),
            nn.Conv3d(base * 16, base * 16, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.BatchNorm3d(base * 16),
            nn.MaxPool3d(kernel_size=3, stride=2),)

        self.avgpool = nn.AdaptiveAvgPool3d((6, 6, 6))
        self.ft = nn.Flatten()
        self.dp1 = nn.Dropout()

        if self.level_node != 0:
            nb_fc0 = base * 16 * 6 * 6 * 6 + 1
        else:
            nb_fc0 = base * 16 * 6 * 6 * 6

        self.ln1 = nn.Linear(nb_fc0, fc1_nodes)
        self.rl1 = nn.ReLU(inplace=True)

        self.dp2 = nn.Dropout()
        self.ln2 = nn.Linear(fc1_nodes, fc2_nodes)
        self.rl2 = nn.ReLU(inplace=True)

        self.dp3 = nn.Dropout()
        self.ln3 = nn.Linear(fc2_nodes, self.num_classes)

    def forward(self, input):  # input would be a tuple of size (1,) if only one element is input
        x = self.features(input)

        return x

dims = 3
batch_x = torch.ones((2, 1, 96, 96, 64))
enc_s_ls = [Cnn2_3dEnc()]  # todo: VGG11
enc_t_name_ls = ['resnet3d_18', 'resnet3d_18', 'resnet3d_34','resnet3d_50', 'resnet3d_101']



class Testkd_loss(unittest.TestCase):
    def test_kd_loss(self):
        for enc_s in enc_s_ls:
            for enc_t_name in enc_t_name_ls:
                kd_loss(batch_x, enc_s, enc_t_name)



if __name__ == "__main__":
    unittest.main()




# -*- coding: utf-8 -*-
# @Time    : 7/4/21 3:14 PM
# @Author  : Jingnan
# @Email   : jiajingnan2222@gmail.com
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
import math
from functools import partial

__all__ = [
    'ResNet', 'resnet10', 'resnet18', 'resnet34', 'resnet50', 'resnet101',
    'resnet152', 'resnet200'
]


def conv3x3x3(in_planes, out_planes, stride=1, dilation=1):
    # 3x3x3 convolution with padding
    return nn.Conv3d(
        in_planes,
        out_planes,
        kernel_size=3,
        dilation=dilation,
        stride=stride,
        padding=dilation,
        bias=False)


def downsample_basic_block(x, planes, stride, no_cuda=False):
    out = F.avg_pool3d(x, kernel_size=1, stride=stride)
    zero_pads = torch.Tensor(
        out.size(0), planes - out.size(1), out.size(2), out.size(3),
        out.size(4)).zero_()
    if not no_cuda:
        if isinstance(out.data, torch.cuda.FloatTensor):
            zero_pads = zero_pads.cuda()

    out = Variable(torch.cat([out.data, zero_pads], dim=1))

    return out


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, dilation=1, downsample=None):
        super(BasicBlock, self).__init__()
        self.conv1 = conv3x3x3(inplanes, planes, stride=stride, dilation=dilation)
        self.bn1 = nn.BatchNorm3d(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3x3(planes, planes, dilation=dilation)
        self.bn2 = nn.BatchNorm3d(planes)
        self.downsample = downsample
        self.stride = stride
        self.dilation = dilation

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)
        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, dilation=1, downsample=None):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv3d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm3d(planes)
        self.conv2 = nn.Conv3d(
            planes, planes, kernel_size=3, stride=stride, dilation=dilation, padding=dilation, bias=False)
        self.bn2 = nn.BatchNorm3d(planes)
        self.conv3 = nn.Conv3d(planes, planes * 4, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm3d(planes * 4)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride
        self.dilation = dilation

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


# class ResNet_with_conv_seg(nn.Module):
#
#     def __init__(self,
#                  block,
#                  layers,
#                  num_seg_classes,
#                  shortcut_type='B',
#                  no_cuda=False):
#         self.inplanes = 64
#         self.no_cuda = no_cuda
#         super(ResNet, self).__init__()
#         self.conv1 = nn.Conv3d(
#             1,
#             64,
#             kernel_size=7,
#             stride=(2, 2, 2),
#             padding=(3, 3, 3),
#             bias=False)
#
#         self.bn1 = nn.BatchNorm3d(64)
#         self.relu = nn.ReLU(inplace=True)
#         self.maxpool = nn.MaxPool3d(kernel_size=(3, 3, 3), stride=2, padding=1)
#         self.layer1 = self._make_layer(block, 64, layers[0], shortcut_type)
#         self.layer2 = self._make_layer(
#             block, 128, layers[1], shortcut_type, stride=2)
#         self.layer3 = self._make_layer(
#             block, 256, layers[2], shortcut_type, stride=1, dilation=2)
#         self.layer4 = self._make_layer(
#             block, 512, layers[3], shortcut_type, stride=1, dilation=4)
#
#         self.conv_seg = nn.Sequential(
#             nn.ConvTranspose3d(
#                 512 * block.expansion,
#                 32,
#                 2,
#                 stride=2
#             ),
#             nn.BatchNorm3d(32),
#             nn.ReLU(inplace=True),
#             nn.Conv3d(
#                 32,
#                 32,
#                 kernel_size=3,
#                 stride=(1, 1, 1),
#                 padding=(1, 1, 1),
#                 bias=False),
#             nn.BatchNorm3d(32),
#             nn.ReLU(inplace=True),
#             nn.Conv3d(
#                 32,
#                 num_seg_classes,
#                 kernel_size=1,
#                 stride=(1, 1, 1),
#                 bias=False)
#         )
#
#         for m in self.modules():
#             if isinstance(m, nn.Conv3d):
#                 m.weight = nn.init.kaiming_normal(m.weight, mode='fan_out')
#             elif isinstance(m, nn.BatchNorm3d):
#                 m.weight.data.fill_(1)
#                 m.bias.data.zero_()
#
#     def _make_layer(self, block, planes, blocks, shortcut_type, stride=1, dilation=1):
#         downsample = None
#         if stride != 1 or self.inplanes != planes * block.expansion:
#             if shortcut_type == 'A':
#                 downsample = partial(
#                     downsample_basic_block,
#                     planes=planes * block.expansion,
#                     stride=stride,
#                     no_cuda=self.no_cuda)
#             else:
#                 downsample = nn.Sequential(
#                     nn.Conv3d(
#                         self.inplanes,
#                         planes * block.expansion,
#                         kernel_size=1,
#                         stride=stride,
#                         bias=False), nn.BatchNorm3d(planes * block.expansion))
#
#         layers = []
#         layers.append(block(self.inplanes, planes, stride=stride, dilation=dilation, downsample=downsample))
#         self.inplanes = planes * block.expansion
#         for i in range(1, blocks):
#             layers.append(block(self.inplanes, planes, dilation=dilation))
#
#         return nn.Sequential(*layers)
#
#     def forward(self, x):
#         x = self.conv1(x)
#         x = self.bn1(x)
#         x = self.relu(x)
#         x = self.maxpool(x)
#         x = self.layer1(x)
#         x = self.layer2(x)
#         x = self.layer3(x)
#         x = self.layer4(x)
#         x = self.conv_seg(x)
#
#         return x


class ResNet(nn.Module):

    def __init__(self,
                 block,
                 layers,
                 num_seg_classes,
                 shortcut_type='B',
                 no_cuda=False):
        self.inplanes = 64
        self.no_cuda = no_cuda
        super(ResNet, self).__init__()
        self.conv1 = nn.Conv3d(
            1,
            64,
            kernel_size=7,
            stride=(2, 2, 2),
            padding=(3, 3, 3),
            bias=False)

        self.bn1 = nn.BatchNorm3d(64)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool3d(kernel_size=(3, 3, 3), stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64, layers[0], shortcut_type)
        self.layer2 = self._make_layer(
            block, 128, layers[1], shortcut_type, stride=2)
        self.layer3 = self._make_layer(
            block, 256, layers[2], shortcut_type, stride=1, dilation=2)
        self.layer4 = self._make_layer(
            block, 512, layers[3], shortcut_type, stride=1, dilation=4)
        #
        # self.conv_seg = nn.Sequential(
        #     nn.ConvTranspose3d(
        #         512 * block.expansion,
        #         32,
        #         2,
        #         stride=2
        #     ),
        #     nn.BatchNorm3d(32),
        #     nn.ReLU(inplace=True),
        #     nn.Conv3d(
        #         32,
        #         32,
        #         kernel_size=3,
        #         stride=(1, 1, 1),
        #         padding=(1, 1, 1),
        #         bias=False),
        #     nn.BatchNorm3d(32),
        #     nn.ReLU(inplace=True),
        #     nn.Conv3d(
        #         32,
        #         num_seg_classes,
        #         kernel_size=1,
        #         stride=(1, 1, 1),
        #         bias=False)
        # )

        for m in self.modules():
            if isinstance(m, nn.Conv3d):
                m.weight = nn.init.kaiming_normal_(m.weight, mode='fan_out')
            elif isinstance(m, nn.BatchNorm3d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()

    def _make_layer(self, block, planes, blocks, shortcut_type, stride=1, dilation=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            if shortcut_type == 'A':
                downsample = partial(
                    downsample_basic_block,
                    planes=planes * block.expansion,
                    stride=stride,
                    no_cuda=self.no_cuda)
            else:
                downsample = nn.Sequential(
                    nn.Conv3d(
                        self.inplanes,
                        planes * block.expansion,
                        kernel_size=1,
                        stride=stride,
                        bias=False), nn.BatchNorm3d(planes * block.expansion))

        layers = []
        layers.append(block(self.inplanes, planes, stride=stride, dilation=dilation, downsample=downsample))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes, dilation=dilation))

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        # x = self.conv_seg(x)

        return x


def resnet10(**kwargs):
    """Constructs a ResNet-18 model.
    """
    model = ResNet(BasicBlock, [1, 1, 1, 1], **kwargs)
    return model


def resnet18(**kwargs):
    """Constructs a ResNet-18 model.
    """
    model = ResNet(BasicBlock, [2, 2, 2, 2], **kwargs)
    return model


def resnet34(**kwargs):
    """Constructs a ResNet-34 model.
    """
    model = ResNet(BasicBlock, [3, 4, 6, 3], **kwargs)
    return model


def resnet50(**kwargs):
    """Constructs a ResNet-50 model.
    """
    model = ResNet(Bottleneck, [3, 4, 6, 3], **kwargs)
    return model


def resnet101(**kwargs):
    """Constructs a ResNet-101 model.
    """
    model = ResNet(Bottleneck, [3, 4, 23, 3], **kwargs)
    return model


def resnet152(**kwargs):
    """Constructs a ResNet-101 model.
    """
    model = ResNet(Bottleneck, [3, 8, 36, 3], **kwargs)
    return model


def resnet200(**kwargs):
    """Constructs a ResNet-101 model.
    """
    model = ResNet(Bottleneck, [3, 24, 36, 3], **kwargs)
    return model



# -*- coding: utf-8 -*-
# @Time    : 6/30/21 5:43 PM
# @Author  : Jingnan
# @Email   : jiajingnan2222@gmail.com

import unittest
import coverage


COV_FLAG = True
if __name__ == "__main__":
    if COV_FLAG:
        cov = coverage.coverage()
        cov.start()

    loader = unittest.TestLoader()
    tests = loader.discover('.', pattern='test_*.py')
    testRunner = unittest.runner.TextTestRunner()
    test_results = testRunner.run(tests)

    if COV_FLAG:
        cov.stop()
        cov.save()
        # 命令行模式展示结果
        cov.report()
        # 生成HTML覆盖率报告
        # cov.html_report(directory='./tests/covhtml')
        cov.xml_report(outfile='cov_report.xml')


    if test_results.wasSuccessful():  # used in github actions to make sure actions fail when tests fails
        exit(0)
    else:
        exit(1)



# -*- coding: utf-8 -*-
# @Time    : 7/26/21 10:49 PM
# @Author  : Jingnan
# @Email   : jiajingnan2222@gmail.com
import torch
import torch.nn as nn
from kd_med.get_resnet3d import generate_model
from kd_med.unet3d import UNet3D
from kd_med.resnet3d import ResNet
import os
from collections import OrderedDict
from google_drive_downloader import GoogleDriveDownloader as gdd


class Option:
    def __init__(self):
        pass


def download_weights_if_needed(weights_dir: str):
    if not os.path.isdir(weights_dir):
        os.makedirs(weights_dir)
    # download weights and unzip weights

    file_id_path_dt = {'1yZF0WHUpri9ajUeJE3wmxqFUW2Pimg7g': os.path.join(weights_dir, 'resnet_10_23dataset.pth'),
                       '1vYvIamI4zWUTD7TGgatxwkMAwqTCpd_B': os.path.join(weights_dir, 'resnet_18_23dataset.pth'),
                       '1estlxOizcJmNTa85zOeyzjQ7-0rzsdK8': os.path.join(weights_dir, 'resnet_34_23dataset.pth'),
                       '18FQtxbbvGWlBXk8bGE7MphV7K0LMMhD0': os.path.join(weights_dir, 'resnet_50_23dataset.pth'),
                       '1izVHemAsKPGu3RvtyzmBJE46C2Qo-OF3': os.path.join(weights_dir, 'resnet_101.pth'),
                       '1yqIIM_PNin8lNeWlxGH_Qs5sjLkPqjSb': os.path.join(weights_dir, 'resnet_152.pth'),
                       '1FWtQqyuOUjf4zk-WbACrj5g4AnuSb_8Z': os.path.join(weights_dir, 'resnet_200.pth'),
                       '192k-iKbvDpnPc1j5Da88Ogrt11JsikVp': os.path.join(weights_dir, 'Genesis_Chest_CT.pt')
                       }
    for id, path in file_id_path_dt.items():
        if not os.path.isfile(path):
            gdd.download_file_from_google_drive(file_id=id, dest_path=path)

    return None


class UNet3DEnc(UNet3D):
    def __init__(self, output_layer = 'out_up_256'):
        super().__init__()
        self.output_layer = output_layer
        self._layers = []
        for l in list(self._modules.keys()):
            self._layers.append(l)
            if l == output_layer:
                break
        self.layers = OrderedDict(zip(self._layers, [getattr(self, l) for l in self._layers]))

    def _forward_impl(self, x):
        for l in self._layers:
            x = self.layers[l](x)
        return x

    def forward(self, x):
        return self._forward_impl(x)


def rename_layers(loaded_dict):
    state_dict = loaded_dict['state_dict']
    unParalled_state_dict = OrderedDict()
    for key in state_dict.keys():
        unParalled_state_dict[key.replace("module.", "")] = state_dict[key]
    return unParalled_state_dict


class PreTrainedEnc:
    weights_dir = os.path.join(os.path.expanduser('~/.cache'), 'kd_med', 'pretrained_weights')
    enc = {}

    @classmethod
    def get(cls, net_name: str):
        if net_name not in cls.enc:
            download_weights_if_needed(cls.weights_dir)
            opt = Option()
            if 'resnet3d' in net_name:
                opt.model = 'resnet'  # do not change it
                opt.model_depth = int(net_name.split('_')[-1])
                if opt.model_depth in [18, 34]:  # https://github.com/Tencent/MedicalNet
                    opt.resnet_shortcut = 'A'
                else:
                    opt.resnet_shortcut = 'B'
                opt.no_cuda = True  # I think it does not matter
                opt.n_seg_classes = 2  # I think it does not matter
                if opt.model_depth in [10, 18, 34, 50]:
                    opt.pretrain_path = cls.weights_dir + "/resnet_" + str(opt.model_depth) + "_23dataset.pth"  # reset if needed
                else:
                    opt.pretrain_path = cls.weights_dir + "/resnet_" + str(opt.model_depth) + ".pth"  # reset if needed

                enc = generate_model(opt)  # generate resnet encoder
                # enc = get_enc(enc, 'conv_seg')
                enc.load_state_dict(rename_layers(torch.load(opt.pretrain_path, map_location=torch.device('cpu'))))

            elif 'resnet2d' in net_name:
                raise NotImplementedError
            elif 'unet3d' == net_name:
                weights_fpath = cls.weights_dir + '/Genesis_Chest_CT.pt'
                base_model = UNet3DEnc()  # Unet encoder

                # Load pre-trained weights
                checkpoint = torch.load(weights_fpath, map_location=torch.device('cpu'))
                # state_dict = checkpoint['state_dict']
                # unParalled_state_dict = {}
                # for key in state_dict.keys():
                #     unParalled_state_dict[key.replace("module.", "")] = state_dict[key]
                base_model.load_state_dict(rename_layers(checkpoint))
                enc = base_model  # encoder
            else:
                raise Exception(f'wrong net name {net_name}')
            cls.enc[net_name] = enc
        return cls.enc[net_name]


# def pretrained_enc(net_name: str):
#     """
#     Availabel net_name:
#     Med3D (https://arxiv.org/pdf/1904.00625.pdf):
#         resnet3d_10, resnet3d_18, resnet3d_34, resnet3d_50, resnet3d_101, resnet3d_152, resnet3d_200
#     Model Genesis (https://arxiv.org/pdf/1908.06912.pdf):
#         unet3d
#         resnet2d_18  # I do not have its pre-trained weights yet, because it is not released.
#
# opt: require following values: model, model_depth, input_W, input_H, input_D,
#     resnet_shortcut, no_cuda, n_seg_classes, phase, pretrain_path
#
#     :param net_name:
#     :return:
#     """
#     # weights_dir = 'pretrained_weights'
#     weights_dir = os.path.join(os.path.expanduser('~/.cache'), 'kd_med', 'pretrained_weights')
#
#     download_weights(weights_dir)
#
#     opt = Option()
#     if 'resnet3d' in net_name:
#         opt.model = 'resnet'  # do not change it
#         opt.model_depth = int(net_name.split('_')[-1])
#         if opt.model_depth in [18, 34]:  # https://github.com/Tencent/MedicalNet
#             opt.resnet_shortcut = 'A'
#         else:
#             opt.resnet_shortcut = 'B'
#         opt.no_cuda = True  # I think it does not matter
#         opt.n_seg_classes = 2  # I think it does not matter
#         if opt.model_depth in [10, 18, 34, 50]:
#             opt.pretrain_path = weights_dir + "/resnet_" + str(opt.model_depth) + "_23dataset.pth"  # reset if needed
#         else:
#             opt.pretrain_path = weights_dir + "/resnet_" + str(opt.model_depth) + ".pth"  # reset if needed
#
#
#         enc = generate_model(opt)  # generate resnet encoder
#         # enc = get_enc(enc, 'conv_seg')
#         enc.load_state_dict(rename_layers(torch.load(opt.pretrain_path)), map_location=torch.device('cpu'))
#
#     elif 'resnet2d' in net_name:
#         raise NotImplementedError
#     elif 'unet3d' == net_name:
#         weights_fpath = weights_dir + '/Genesis_Chest_CT.pt'
#         base_model = UNet3DEnc()  # Unet encoder
#
#         # Load pre-trained weights
#         checkpoint = torch.load(weights_fpath)
#         # state_dict = checkpoint['state_dict']
#         # unParalled_state_dict = {}
#         # for key in state_dict.keys():
#         #     unParalled_state_dict[key.replace("module.", "")] = state_dict[key]
#         base_model.load_state_dict(rename_layers(checkpoint), map_location=torch.device('cpu'))
#         enc = base_model  # encoder
#     else:
#         raise Exception(f'wrong net name {net_name}')
#     return enc




# -*- coding: utf-8 -*-
# @Time    : 7/27/21 5:23 PM
# @Author  : Jingnan
# @Email   : jiajingnan2222@gmail.com
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="kd_med", # Replace with your own username
    version="0.0.7",
    author="Jingnan",
    author_email="jiajingnan2222@gmail.com",
    description="A plug-in for knowledge distillation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jingnan-jia/kd_med",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=['googledrivedownloader', 'torch'],

)



# -*- coding: utf-8 -*-
# @Time    : 7/27/21 7:57 PM
# @Author  : Jingnan
# @Email   : jiajingnan2222@gmail.com
import unittest
import tempfile
import os
import torch
import torch.nn as nn

from parameterized import parameterized
from kd_med.kd_loss import GetEncSConv

import numpy as np

TEST_CASE_3d_1 = [nn.Sequential(nn.Conv3d(1, 48, kernel_size=(3, 3, 3), stride=(2, 2, 2), padding=(3, 3, 3)),
                                nn.Conv3d(48, 48, kernel_size=(3, 3, 3), stride=(2, 2, 2), padding=(3, 3, 3))),
                  nn.Conv3d(1, 32, kernel_size=(3, 3, 3), stride=(2, 2, 2), padding=(3, 3, 3)),
                  torch.ones((2, 1, 64, 64, 32)),
                  3]

TEST_CASE_3d_2 = [nn.Sequential(nn.Conv3d(1, 48, kernel_size=(3, 3, 3), stride=(2, 2, 2), padding=(3, 3, 3)),
                                nn.MaxPool3d(kernel_size=(2, 2, 2)),
                                nn.Conv3d(48, 18, kernel_size=(3, 3, 3), stride=(2, 2, 2), padding=(3, 3, 3))),
                  nn.Conv3d(1, 32, kernel_size=(3, 3, 3), stride=(2, 2, 2), padding=(3, 3, 3)),
                  torch.ones((2, 1, 32, 16, 8)),
                  3]

TEST_CASE_2d = [nn.Sequential(nn.Conv2d(1, 64, kernel_size=(3, 3), stride=(2, 2), padding=(3, 3)),
                              nn.Conv2d(64, 64, kernel_size=(3, 3), stride=(2, 2), padding=(3, 3))),
                nn.Conv2d(1, 16, kernel_size=(3, 3), stride=(2, 2), padding=(3, 3)),
                torch.ones((2, 1, 64, 32)),
                2]


class TestGetEncSConv(unittest.TestCase):
    @parameterized.expand([TEST_CASE_3d_1, TEST_CASE_3d_2, TEST_CASE_2d])
    def test_GetEncSConv(self, enc_t, enc_s, input_img, dims):
        enc_s_and_conv1 = GetEncSConv().get(enc_t, enc_s, dims)
        enc_s_and_conv2 = GetEncSConv().get(enc_t, enc_s, dims)
        enc_s_and_conv3 = GetEncSConv().get(enc_t, enc_s, dims)
        # print(enc_s_and_conv1)

        self.assertIs(enc_s_and_conv1, enc_s_and_conv2)
        self.assertIs(enc_s_and_conv1, enc_s_and_conv3)

        self.assertEqual(enc_s_and_conv1(input_img).shape, enc_t(input_img).shape)


if __name__ == "__main__":
    unittest.main()
